@@ -1,8 +1,9 @@
 __author__ = "Johannes Köster"
-__copyright__ = "Copyright 2021, Johannes Köster"
+__copyright__ = "Copyright 2022, Johannes Köster"
 __email__ = "johannes.koester@uni-due.de"
 __license__ = "MIT"
 
+from tempfile import TemporaryFile
 import tokenize
 import textwrap
 import os
@@ -161,11 +162,11 @@ class KeywordState(TokenAutomaton):
     def is_block_end(self, token):
         return (self.line and self.indent <= 0) or is_eof(token)
 
-    def block(self, token):
+    def block(self, token, force_block_end=False):
         if self.lasttoken == "\n" and is_comment(token):
             # ignore lines containing only comments
             self.line -= 1
-        if self.is_block_end(token):
+        if force_block_end or self.is_block_end(token):
             yield from self.decorate_end(token)
             yield "\n", token
             raise StopAutomaton(token)
@@ -426,6 +427,10 @@ class Threads(RuleKeywordState):
     pass
 
 
+class Retries(RuleKeywordState):
+    pass
+
+
 class Shadow(RuleKeywordState):
     pass
 
@@ -648,6 +653,14 @@ class Wrapper(Script):
         )
 
 
+class TemplateEngine(Script):
+    start_func = "@workflow.template_engine"
+    end_func = "render_template"
+
+    def args(self):
+        yield (", input, output, params, wildcards, config, rule")
+
+
 class CWL(Script):
     start_func = "@workflow.cwl"
     end_func = "cwl"
@@ -666,6 +679,7 @@ rule_property_subautomata = dict(
     params=Params,
     threads=Threads,
     resources=Resources,
+    retries=Retries,
     priority=Priority,
     version=Version,
     log=Log,
@@ -692,6 +706,7 @@ class Rule(GlobalKeywordState):
         script=Script,
         notebook=Notebook,
         wrapper=Wrapper,
+        template_engine=TemplateEngine,
         cwl=CWL,
         **rule_property_subautomata,
     )
@@ -749,6 +764,8 @@ class Rule(GlobalKeywordState):
                     or token.string == "shell"
                     or token.string == "script"
                     or token.string == "wrapper"
+                    or token.string == "notebook"
+                    or token.string == "template_engine"
                     or token.string == "cwl"
                 ):
                     if self.run:
@@ -762,7 +779,7 @@ class Rule(GlobalKeywordState):
                 elif self.run:
                     raise self.error(
                         "No rule keywords allowed after "
-                        "run/shell/script/wrapper/cwl in "
+                        "run/shell/script/notebook/wrapper/template_engine/cwl in "
                         "rule {}.".format(self.rulename),
                         token,
                     )
@@ -943,11 +960,12 @@ class UseRule(GlobalKeywordState):
         )
         yield "\n"
 
-        # yield with block
-        yield from self._with_block
+        if self._with_block:
+            # yield with block
+            yield from self._with_block
 
-        yield "@workflow.run"
-        yield "\n"
+            yield "@workflow.run"
+            yield "\n"
 
         rulename = self.rules[0]
         if rulename == "*":
@@ -1051,7 +1069,7 @@ class UseRule(GlobalKeywordState):
                 )
         elif is_newline(token) or is_comment(token) or is_eof(token):
             # end of the statement, close block manually
-            yield from self.block(token)
+            yield from self.block(token, force_block_end=True)
         else:
             self.error(
                 "Expecting either 'as', 'with' or end of line in 'use rule' statement.",
@@ -1080,7 +1098,7 @@ class UseRule(GlobalKeywordState):
             yield from ()
         elif is_newline(token) or is_comment(token) or is_eof(token):
             # end of the statement, close block manually
-            yield from self.block(token)
+            yield from self.block(token, force_block_end=True)
         else:
             self.error(
                 "Expecting rulename modifying pattern (e.g. modulename_*) after 'as' keyword.",
@@ -1111,7 +1129,7 @@ class UseRule(GlobalKeywordState):
                 )
             except StopAutomaton as e:
                 self.indentation(e.token)
-                self.block(e.token)
+                yield from self.block(e.token)
         else:
             self.error(
                 "Expecting a keyword or comment "
