@@ -1,5 +1,5 @@
 __author__ = "Johannes Köster"
-__copyright__ = "Copyright 2021, Johannes Köster"
+__copyright__ = "Copyright 2022, Johannes Köster"
 __email__ = "johannes.koester@uni-due.de"
 __license__ = "MIT"
 
@@ -15,6 +15,7 @@ from functools import lru_cache, partial
 from itertools import filterfalse, count
 from pathlib import Path
 
+import snakemake.exceptions
 from snakemake.logging import logger
 from snakemake.jobs import jobfiles
 from snakemake.utils import listfiles
@@ -167,7 +168,7 @@ class Persistence:
 
     def lock(self):
         if self.locked:
-            raise IOError("Another snakemake process " "has locked this directory.")
+            raise snakemake.exceptions.LockException()
         self._lock(self.all_inputfiles(), "input")
         self._lock(self.all_outputfiles(), "output")
 
@@ -186,7 +187,7 @@ class Persistence:
         shutil.rmtree(self._lockdir)
 
     def cleanup_metadata(self, path):
-        self._delete_record(self._metadata_path, path)
+        return self._delete_record(self._metadata_path, path)
 
     def cleanup_shadow(self):
         if os.path.exists(self.shadow_path):
@@ -240,6 +241,9 @@ class Persistence:
                     "starttime", None
                 )
             endtime = f.mtime.local_or_remote() if f.exists else fallback_time
+
+            checksums = ((infile, infile.checksum()) for infile in job.input)
+
             self._record(
                 self._metadata_path,
                 {
@@ -256,6 +260,11 @@ class Persistence:
                     "job_hash": hash(job),
                     "conda_env": conda_env,
                     "container_img_url": job.container_img_url,
+                    "input_checksums": {
+                        infile: checksum
+                        for infile, checksum in checksums
+                        if checksum is not None
+                    },
                 },
                 f,
             )
@@ -322,37 +331,74 @@ class Persistence:
     def code(self, path):
         return self.metadata(path).get("code")
 
+    def conda_env(self, path):
+        return self.metadata(path).get("conda_env")
+
+    def container_img_url(self, path):
+        return self.metadata(path).get("container_img_url")
+
+    def input_checksums(self, job, input_path):
+        """Return all checksums of the given input file
+        recorded for the output of the given job.
+        """
+        return set(
+            self.metadata(output_path).get("input_checksums", {}).get(input_path)
+            for output_path in job.output
+        )
+
     def version_changed(self, job, file=None):
-        """Yields output files with changed versions of bool if file given."""
+        """Yields output files with changed versions or bool if file given."""
         return _bool_or_gen(self._version_changed, job, file=file)
 
     def code_changed(self, job, file=None):
-        """Yields output files with changed code of bool if file given."""
+        """Yields output files with changed code or bool if file given."""
         return _bool_or_gen(self._code_changed, job, file=file)
 
     def input_changed(self, job, file=None):
-        """Yields output files with changed input of bool if file given."""
+        """Yields output files with changed input or bool if file given."""
         return _bool_or_gen(self._input_changed, job, file=file)
 
     def params_changed(self, job, file=None):
-        """Yields output files with changed params of bool if file given."""
+        """Yields output files with changed params or bool if file given."""
         return _bool_or_gen(self._params_changed, job, file=file)
 
+    def conda_env_changed(self, job, file=None):
+        """Yields output files with changed conda env or bool if file given."""
+        return _bool_or_gen(self._conda_env_changed, job, file=file)
+
+    def container_changed(self, job, file=None):
+        """Yields output files with changed container img or bool if file given."""
+        return _bool_or_gen(self._container_changed, job, file=file)
+
     def _version_changed(self, job, file=None):
         assert file is not None
-        return self.version(file) != job.rule.version
+        recorded = self.version(file)
+        return recorded is not None and recorded != job.rule.version
 
     def _code_changed(self, job, file=None):
         assert file is not None
-        return self.code(file) != self._code(job.rule)
+        recorded = self.code(file)
+        return recorded is not None and recorded != self._code(job.rule)
 
     def _input_changed(self, job, file=None):
         assert file is not None
-        return self.input(file) != self._input(job)
+        recorded = self.input(file)
+        return recorded is not None and recorded != self._input(job)
 
     def _params_changed(self, job, file=None):
         assert file is not None
-        return self.params(file) != self._params(job)
+        recorded = self.params(file)
+        return recorded is not None and recorded != self._params(job)
+
+    def _conda_env_changed(self, job, file=None):
+        assert file is not None
+        recorded = self.conda_env(file)
+        return recorded is not None and recorded != self._conda_env(job)
+
+    def _container_changed(self, job, file=None):
+        assert file is not None
+        recorded = self.container_img_url(file)
+        return recorded is not None and recorded != job.container_img_url
 
     def noop(self, *args):
         pass
@@ -399,9 +445,14 @@ class Persistence:
             recdirs = os.path.relpath(os.path.dirname(recpath), start=subject)
             if recdirs != ".":
                 os.removedirs(recdirs)
+            return True
         except OSError as e:
-            if e.errno != 2:  # not missing
+            if e.errno != 2:
+                # not missing
                 raise e
+            else:
+                # file is missing, report failure
+                return False
 
     @lru_cache()
     def _read_record_cached(self, subject, id):
@@ -451,7 +502,7 @@ class Persistence:
 
     def all_outputfiles(self):
         # we only look at output files that will be updated
-        return jobfiles(self.dag.needrun_jobs, "output")
+        return jobfiles(self.dag.needrun_jobs(), "output")
 
     def all_inputfiles(self):
         # we consider all input files, also of not running jobs
