@@ -66,11 +66,12 @@ class DefaultResources:
                         ):
                             # Missing input files are handled by the caller
                             raise WorkflowError(
-                                "Failed to evaluate DefaultResources value "
+                                "Failed to evaluate default resources value "
                                 "'{}'.\n"
                                 "    String arguments may need additional "
                                 "quoting. Ex: --default-resources "
-                                "\"tmpdir='/home/user/tmp'\".".format(val)
+                                "\"tmpdir='/home/user/tmp'\".".format(val),
+                                e,
                             )
                         raise e
                     return value
